from typing import Any, Callable, Optional, Type, Union, overload

from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db.models import Model
from django.db.models.constants import LOOKUP_SEP
from django.forms import BaseForm, BaseModelForm, CharField, Form, ModelForm, modelform_factory

from django_json_model_field.db.models import JSONModel

from .nested_form_field import NestedFormField

AnyModel = Union[Model, JSONModel]
NestedModelClassFn = Callable[[Union[Model, dict]], Optional[Type[AnyModel]]]
NestedModelClassArg = Union[Type[AnyModel], NestedModelClassFn]


class _SelectorOnlyForm(Form):
    """
    Form used internally by NestedModelFormField to track the value of the selector field when there is no form required
    for the current selector value.
    """

    _selector_value = CharField(widget=CharField.hidden_widget, required=False)


class BaseSelectedNestedModelForm(ModelForm):
    """
    Base ModelForm class used internally by NestedModelFormField to track the value of the selector field when a
    selector is used to vary the model class used for the form.
    """

    _selector_value = CharField(widget=CharField.hidden_widget, required=False)


class NestedModelFormField(NestedFormField):
    """
    Subclass of `NestedFormField` that adds support for generating forms based on JSONModel or Model classes, as well
    as multi-variant models using a "selector" field.
    """

    _selector_value_field = "_selector_value"
    _selector_value_changed = "selector_value_changed"

    default_error_messages = {
        # hide the error message for model_validation so that a validation errors on the nested form do not cause an
        # additional, more generic validation error message to be shown for this field
        "model_validation": ""
    }

    @overload
    def __init__(self, nested_model_class: Type[JSONModel], *args, **kwargs):
        ...

    @overload
    def __init__(
        self,
        nested_model_class: NestedModelClassFn,
        selector_field: str,
        *args,
        **kwargs,
    ):
        ...

    def __init__(
        self,
        nested_model_class: NestedModelClassArg,
        selector_field: str = None,
        get_selector_value: Callable[[Model], Optional[Any]] = None,
        *args,
        **kwargs,
    ):
        self.selector_field = selector_field
        if self.selector_field:
            # this must happen before super().__init__ because it deals with instantiating the widget
            self.widget = self.widget.with_fields_from_host(self.selector_field)

        super().__init__(self._get_nested_model_form_class, *args, **kwargs)

        if isinstance(nested_model_class, type) and issubclass(nested_model_class, JSONModel):
            self.nested_model_class = nested_model_class
        else:
            setattr(self, "_get_nested_model_class", nested_model_class)

        if get_selector_value:
            self._get_selector_value_from_instance = get_selector_value

    def get_nested_form_kwargs(self, host_form: BaseForm, field_name: str, nested_form_class: Type[BaseForm]) -> dict:
        # the "initial" for JSONModelFields can be a JSONModel instance - move it to the "instance" kwarg
        # FIXME: why?
        instance = self._fix_nested_form_initial(host_form, field_name)

        kwargs = super().get_nested_form_kwargs(host_form, field_name, nested_form_class)
        if instance:
            kwargs.update(instance=instance)
        return kwargs

    def _fix_nested_form_initial(self, host_form: BaseForm, field_name: str) -> Optional[JSONModel]:
        initial = host_form.initial.pop(field_name, {})
        instance: Optional[JSONModel] = None

        if isinstance(initial, JSONModel):
            instance = initial
            initial = {}

        host_form.initial[field_name] = initial
        return instance

    def get_nested_form_initial(self, host_form: BaseForm, field_name: str, nested_form_class: Type[BaseForm]) -> dict:
        initial = super().get_nested_form_initial(host_form, field_name, nested_form_class)

        if self.selector_field and isinstance(host_form, BaseModelForm):
            initial = initial or {}
            initial.update({self._selector_value_field: self._get_selector_value(host_form)})

        return initial

    def get_nested_form_data(self, host_form: BaseForm, field_name: str, nested_form_class: Type[BaseForm]) -> dict:
        data = super().get_nested_form_data(host_form, field_name, nested_form_class)

        if data is None:
            return None

        if self.selector_field:
            self._update_form_data_for_selector_field(host_form, field_name, nested_form_class, data)

        return data

    def _update_form_data_for_selector_field(
        self, host_form: BaseForm, field_name: str, nested_form_class: Type[BaseForm], data: dict
    ):
        """
        Sets the hidden selector value field's value based on the value selected the last time the form was
        submitted. This allows the condition that triggers the "selector_value_changed" ValidationError in
        NestedModelFormField.validate to be cleared so that the user can save their changes on the next
        request.
        """

        selector_value = self._get_selector_value(host_form)
        if selector_value is not None:

            selector_data_key = host_form.add_prefix(f"{field_name}-{self._selector_value_field}")
            data[selector_data_key] = selector_value

            if host_form.is_bound and self.selector_field in host_form.changed_data:
                self._update_form_data_for_selector_field_change(host_form, field_name, nested_form_class, data)

    def _update_form_data_for_selector_field_change(
        self, host_form: BaseForm, field_name: str, nested_form_class: Type[BaseForm], data: dict
    ):
        """
        If the value of the selector field changes, set initial data using defaults for the new form.
        Since the host form is already bound, any "initial" data provided for the nested form will be ignored,
        preventing defaults from being set as they usually are.
        """

        # TODO: data for field names shared between the old form and the new one should get removed to avoid
        #       pre-populating the field on the new form with the old data

        # TODO: known issue - when changing the selector on an existing object from the original type, to a different
        #       selector value that uses a different model, and then back causes the original data to be lost.
        #       The missing data can probably be repopulated from the original instance.

        fields = nested_form_class._meta.model._meta.fields
        data.update(
            {
                host_form.add_prefix(f"{field_name}-{nested_field.name}"): nested_field.get_default()
                for nested_field in fields
                if nested_field.has_default()
            }
        )

    def validate(self, value):
        validation_form = self.get_form_for_validation(cleaned_data=value)
        if validation_form is not None and self.selector_field is not None and LOOKUP_SEP not in self.selector_field:
            # FIXME: remove LOOKUP_SEP not in self.selector_field and implement a check if the relation is editable
            validation_form.is_valid()  # trigger validation / change detection
            initial_selector_value = validation_form.cleaned_data.get(self._selector_value_field)
            instance = getattr(validation_form, "instance", None)
            required_selector_value = getattr(instance, self._selector_value_field, None)
            if (
                initial_selector_value or required_selector_value
            ) and initial_selector_value != required_selector_value:
                # if the select field has changed, raise a ValidationError to prevent the user from unintentionally
                # saving default values for the new form.
                raise ValidationError(
                    "The '{selector_field}' field value has changed, "
                    "please verify entries in the updated form and save.".format(selector_field=self.selector_field),
                    code=self._selector_value_changed,
                )

        # this must happen AFTER the above check, otherwise validation errors for the new form may be thrown first
        super().validate(value)

    def _get_nested_model_class(self, host_instance: Union[AnyModel, dict]) -> Optional[Type[AnyModel]]:
        return self.nested_model_class

    def get_nested_model_class(self, host: Optional[Union[BaseForm, dict]]) -> Optional[Type[AnyModel]]:
        if hasattr(self, "nested_model_class"):
            return self.nested_model_class

        if host is None:
            return None

        if isinstance(host, BaseModelForm):
            return self._get_nested_model_class(host.instance)

        return self._get_nested_model_class(host)

    def _get_nested_model_form_class(
        self, host: Optional[Union[BaseForm, dict]], raw_data: dict = None
    ) -> Optional[BaseForm]:
        """
        Returns a form type used to represent the field's data
        """

        if host == {} and LOOKUP_SEP in self.selector_field and raw_data:
            # for selectors using a relation field, inject the selector value to allow the correct model class to
            # get looked up, since it will be the same as when the form was initially rendered
            host.update({self.selector_field: raw_data.get(self._selector_value_field)})

        model_class = self.get_nested_model_class(host)

        if model_class is None and self.selector_field is None:
            # this shouldn't ever happen - it would've gotten caught by the model and field checks, but having the
            # conditional check here helps make mypy happy
            raise ImproperlyConfigured("get_nested_model_class returned None with no selector field")

        fields = [field.name for field in model_class._meta.fields] if model_class else []

        if self.selector_field is None:
            # simplest use case: static model class, don't need to worry about any selector data
            return modelform_factory(model_class, fields=fields)

        if model_class is not None:
            # next simplest use case: variable model class, requires using BaseSelectedNestedModelForm to track
            # changes to the selector field
            fields.append(self._selector_value_field)
            return (
                modelform_factory(model_class, form=BaseSelectedNestedModelForm, fields=fields) if model_class else None
            )

        # other variable model class use cases when no model_class is returned
        selector_value = self._get_selector_value(host)
        if selector_value is None:
            # either it's a new object (newly initialized by the ModelForm), or the selector field is optional
            # and doesn't have a value - _SelectorOnlyForm allows tracking the selector field when no other data is
            # used
            return _SelectorOnlyForm

        # selector field has a value, but it doesn't map to a model (does not require additional data)
        return None

    def _get_selector_value(self, host: Optional[Union[BaseForm, dict]]) -> Optional[Any]:
        if host is None:
            return None

        if isinstance(host, BaseModelForm) and self.selector_field:
            if LOOKUP_SEP in self.selector_field:
                return self._get_selector_value_from_instance(host.instance)

            selector_field = host.fields[self.selector_field]
            selector_value = self._get_selector_value_from_instance(host.instance)
            return None if selector_value in selector_field.empty_values else selector_value

        selector_value = host.get(self.selector_field)
        # TODO: using self.empty_values since there's no reference to the model field ... it's probably good enough?
        return None if selector_value in self.empty_values else selector_value

    def _get_selector_value_from_instance(self, host: Model) -> Optional[Any]:
        return getattr(host, self.selector_field)



from .json_model_decoder import JSONModelDecoder
from .json_model_encoder import JSONModelEncoder, collect_data
from .types import JSONClassWrapper, JSONModelDict

dumps = JSONModelEncoder().encode
loads = JSONModelDecoder().decode



import inspect
from collections import OrderedDict
from typing import TYPE_CHECKING, Type

from django.core.exceptions import FieldDoesNotExist
from django.db.models import Field
from django.db.models.options import Options as ModelOptions

from django.utils.datastructures import ImmutableList
from django.utils.functional import cached_property
from django.utils.translation import override

if TYPE_CHECKING:
    from .json_model import JSONModel

PROXY_PARENTS = object()

IMMUTABLE_WARNING = (
    "The return type of '%s' should never be mutated. If you want to manipulate this list "
    "for your own use, make a copy first."
)

DEFAULT_NAMES = (
    "abstract",
    "ordering",
    "verbose_name",
    "verbose_name_plural",
)


def make_immutable_fields_list(name, data):
    return ImmutableList(data, warning=IMMUTABLE_WARNING % name)


class JSONModelOptions:
    """
    Options class for JSONModel

    Adapted from Django's Options class (imported in this module as ModelOptions for easier disambiguation)
    """

    FORWARD_PROPERTIES = {"fields", "concrete_fields", "local_concrete_fields", "_forward_fields_map"}
    REVERSE_PROPERTIES = {"related_objects", "fields_map", "_relation_tree"}

    model: Type["JSONModel"]

    # not supported
    proxy = False
    proxy_for_model = None

    def __init__(self, meta):
        self._get_fields_cache = {}
        self.local_fields = []
        self.private_fields = []
        self.model_name = None
        self.verbose_name = None
        self.verbose_name_plural = None
        self.object_name = None
        self.meta = meta
        self.parents = OrderedDict()
        self.concrete_model = None
        self.ordering = []
        self._ordering_clash = False
        self.abstract = False

        ##########################################################################################################
        # attributes used for compatibility with Model; values are only used for conditional checks an enumeration
        ##########################################################################################################

        self.app_label = None
        self.db_table = "n/a"
        self.required_db_features: tuple = ()
        self.local_many_to_many: list = []
        self.many_to_many: list = []
        self.proxy = None
        self._relation_tree = ()

    def get_field(self, field_name):
        """
        Return a field instance given the name of a forward or reverse field.
        """

        # Adapted from Options.get_field to omit checks for related fields, which are not supported

        try:
            # Retrieve field instance by name from cached or just-computed
            # field map.
            return self.fields_map[field_name]
        except KeyError:
            raise FieldDoesNotExist("%s has no field named '%s'" % (self.object_name, field_name))

    def __repr__(self):
        return "<JSONModelOptions for %s>" % self.object_name

    def __str__(self):
        return self.model_name

    ###################################################################################################################
    # passthrough methods - these can be used as-is from Django's model Options class directly rather than copy/pasting
    # the code
    ###################################################################################################################

    def add_field(self, field, private=False):
        if field.is_relation:
            raise TypeError("Relation fields are not supported")

        return ModelOptions.add_field(self, field, private)

    def contribute_to_class(self, cls, name):
        return ModelOptions.contribute_to_class(self, cls, name)

    def get_base_chain(self, model):
        return ModelOptions.get_base_chain(self, model)

    def get_parent_list(self):
        return ModelOptions.get_parent_list(self)

    def get_ancestor_link(self, ancestor):
        return ModelOptions.get_ancestor_link(self, ancestor)

    def get_fields(self, include_parents=True, include_hidden=False):
        return ModelOptions.get_fields(self, include_parents, include_hidden)

    def _get_fields(self, forward=True, reverse=True, include_parents=True, include_hidden=False, seen_models=None):
        return ModelOptions._get_fields(self, forward, reverse, include_parents, include_hidden, seen_models)

    def _expire_cache(self, forward=True, reverse=True):
        ModelOptions._expire_cache(self, forward, reverse)

    ####################################################################################################################
    # property re-implementations - these are copied from Django's model Options class since they cannot be accessed via
    #                               a passthrough call like the methods due to being properties
    ####################################################################################################################

    @cached_property
    def _property_names(self):
        """Return a set of the names of the properties defined on the model."""
        names = []
        for name in dir(self.model):
            attr = inspect.getattr_static(self.model, name)
            if isinstance(attr, property):
                names.append(name)
        return frozenset(names)

    @property
    def verbose_name_raw(self):
        """Return the untranslated verbose name."""
        with override(None):
            return str(self.verbose_name)

    @cached_property
    def fields(self) -> list[Field]:
        """
        Return a list of all forward fields on the model and its parents,
        excluding ManyToManyFields.

        Private API intended only to be used by Django itself; get_fields()
        combined with filtering of field properties is the public API for
        obtaining this field list.
        """
        # For legacy reasons, the fields property should only contain forward
        # fields that are not private or with a m2m cardinality. Therefore we
        # pass these three filters as filters to the generator.
        # The third lambda is a longwinded way of checking f.related_model - we don't
        # use that property directly because related_model is a cached property,
        # and all the models may not have been loaded yet; we don't want to cache
        # the string reference to the related_model.
        def is_not_an_m2m_field(f):
            return not (f.is_relation and f.many_to_many)

        def is_not_a_generic_relation(f):
            return not (f.is_relation and f.one_to_many)

        def is_not_a_generic_foreign_key(f):
            return not (
                f.is_relation and f.many_to_one and not (hasattr(f.remote_field, "model") and f.remote_field.model)
            )

        return make_immutable_fields_list(
            "fields",
            (
                f
                for f in self._get_fields(reverse=False)
                if is_not_an_m2m_field(f) and is_not_a_generic_relation(f) and is_not_a_generic_foreign_key(f)
            ),
        )

    @cached_property
    def concrete_fields(self):
        """
        Return a list of all concrete fields on the model and its parents.

        Private API intended only to be used by Django itself; get_fields()
        combined with filtering of field properties is the public API for
        obtaining this field list.
        """
        return make_immutable_fields_list("concrete_fields", (f for f in self.fields if f.concrete))

    @cached_property
    def local_concrete_fields(self):
        """
        Return a list of all concrete fields on the model.

        Private API intended only to be used by Django itself; get_fields()
        combined with filtering of field properties is the public API for
        obtaining this field list.
        """
        return make_immutable_fields_list("local_concrete_fields", (f for f in self.local_fields if f.concrete))

    @cached_property
    def _forward_fields_map(self):
        res = {}
        fields = self._get_fields(reverse=False)
        for field in fields:
            res[field.name] = field
            # Due to the way Django's internals work, get_field() should also
            # be able to fetch a field by attname. In the case of a concrete
            # field with relation, includes the *_id name too
            try:
                res[field.attname] = field
            except AttributeError:
                pass
        return res

    @cached_property
    def fields_map(self):
        res = {}
        fields = self._get_fields(forward=False, include_hidden=True)
        for field in fields:
            res[field.name] = field
            # Due to the way Django's internals work, get_field() should also
            # be able to fetch a field by attname. In the case of a concrete
            # field with relation, includes the *_id name too
            try:
                res[field.attname] = field
            except AttributeError:
                pass
        return res

    ###################################################################################################################
    # placeholder methods - these are methods used by Django Models that pertain to functionality not needed or not
    #                       supported by JSONModel. They are provided as noops so to allow JSONModel classes to be used
    #                       interchangeably with Django Model classes for ModelForms and ModelAdmin.
    ###################################################################################################################

    def setup_pk(self, field):
        pass



from __future__ import annotations

from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class JSONModelDict(Protocol):

    def as_db_json(self) -> JSONClassWrapper | dict[str, Any]:
        ...


class JSONClassWrapper:
    """
    Wrapper class to allow classes that are natively JSON serializable to be deserialized to a different class.

    Must be used with JSONModelEncoder and JSONModelDecoder.

    This is helpful for subclasses of dicts that would otherwise be deserialized to a standard dict.
    """

    def __init__(self, wraps: type, data: dict[str, Any]):
        self.data = data
        self.wraps = wraps



from typing import Optional, TYPE_CHECKING, Type, Union

from django.core import checks
from django.db.models import Model

from .base_json_model_field import BaseJSONModelField

if TYPE_CHECKING:
    from .json_model import JSONModel


class JSONModelField(BaseJSONModelField):

    def __init__(
        self,
        json_model: Optional[Type["JSONModel"]],
        verbose_name: str = None,
        name: str = None,
        null: bool = True,
        blank: bool = True,
        encoder=None,
        decoder=None,
        **kwargs,
    ):
        """

        Parameters
        ----------
        json_model     A single JSONModel class that will be used to work with and validate JSON data stored using the
                       field.
        """
        # validation of model field arguments is handled by Django's checks system - see the `check` method below.

        self.json_model = json_model
        super().__init__(verbose_name, name, encoder, decoder, null=null, blank=blank, **kwargs)

    def get_formfield_kwargs(self, **kwargs):
        kwargs.update(
            nested_model_class=self.json_model,
        )
        return kwargs

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        args = [self.json_model, *args]
        return name, path, args, kwargs

    def get_json_model_class(self, host: Union[Model, dict]) -> Optional[Type["JSONModel"]]:
        """
        Returns the `JSONModel` class to be used to represent the field data
        """

        return self.json_model

    def _check_json_model_arguments(self, **kwargs):
        """
        Checks arguments related to specifying the JSONModel for the field to ensure the right combination of arguments
        was provided, and that the arguments are of valid types.
        """

        if self.json_model is None:
            return [checks.Critical("json_model must have a value", obj=self)]

        return self._check_model_argument(self.json_model, "json_model")

    def _check_json_models(self, **kwargs):
        """
        Runs model checks on any JSONModel types specified for use in the field.
        """

        if self.json_model:
            return self.json_model.check(**kwargs)

        return []



from .nested_form_field import NestedFormClassArg, NestedFormClassFn, NestedFormField
from .nested_model_form_field import NestedModelClassArg, NestedModelClassFn, NestedModelFormField
from .widgets import NestedFormInput



import copy
from itertools import chain
from typing import Any, Protocol, runtime_checkable

from django.core.exceptions import FieldDoesNotExist, FieldError
from django.db.models import Model
from django.db.models.base import ModelBase

from .json_model_options import JSONModelOptions


@runtime_checkable
class ContributesToClass(Protocol):

    def contribute_to_class(self, cls: type, name: str, private_only: bool = False):
        ...


class JSONModelBase(type):
    """
    Metaclass for all JSON Models.

    Adapted from Django's ModelBase class.

    Django uses metaclasses to accomplish some of the "magic" functionality related to fields. This is how fields can
    be defined as class attributes, but function as properties on an instance of the model class.

    The metaclass is used to inspect the way the model is defined, including collecting declaration data from super-
    classes (not currently supported for JSONModel). It actually creates a new type of the same name that is used for
    creating instances of the model. Fields implement contribute_to_class in order to control how their values are
    accessed from model instances.

    This information is then put into an instance of the "options" class, which is set as the _meta attribute of the
    final model type.
    """

    def __new__(mcs, name, bases, attrs, **kwargs):
        # Adapted from ModelBase.__new__ to omit logic related to subclassing / abstract classes, relation fields, and
        # proxy models.

        super_new = super().__new__

        # Also ensure initialization is only performed for subclasses of Model
        # (excluding Model class itself).
        parents = [b for b in bases if isinstance(b, JSONModelBase)]
        if not parents:
            return super_new(mcs, name, bases, attrs)

        # Create the class.
        module = attrs.pop("__module__")
        new_attrs = {"__module__": module}
        classcell = attrs.pop("__classcell__", None)
        if classcell is not None:
            new_attrs["__classcell__"] = classcell
        attr_meta = attrs.pop("Meta", None)
        # Pass all attrs without a (Django-specific) contribute_to_class()
        # method to type.__new__() so that they're properly initialized
        # (i.e. __set_name__()).
        contributable_attrs = {}
        for obj_name, obj in list(attrs.items()):
            if isinstance(obj, ContributesToClass):
                contributable_attrs[obj_name] = obj
            else:
                new_attrs[obj_name] = obj
        new_class = super_new(mcs, name, bases, new_attrs, **kwargs)

        abstract = getattr(attr_meta, "abstract", False)
        meta = attr_meta or getattr(new_class, "Meta", None)
        base_meta = getattr(new_class, "_meta", None)

        # skipping ModelBase logic related to app_label since it's not used (Django uses it for its model registry)

        new_class.add_to_class("_meta", JSONModelOptions(meta))
        if not abstract and base_meta and not base_meta.abstract:
            # Non-abstract child classes inherit some attributes from their
            # non-abstract parent (unless an ABC comes before it in the
            # method resolution order).
            if not hasattr(meta, "ordering"):
                new_class._meta.ordering = base_meta.ordering

        # Add remaining attributes (those with a contribute_to_class() method)
        # to the class.
        for obj_name, obj in contributable_attrs.items():
            new_class.add_to_class(obj_name, obj)

        # All the fields of any type declared on this model
        new_fields = chain(
            new_class._meta.local_fields, new_class._meta.local_many_to_many, new_class._meta.private_fields
        )
        field_names = {f.name for f in new_fields}

        # skipping ModelBase logic related to proxying
        new_class._meta.concrete_model = new_class

        # skipping ModelBase logic related to parent links (relations)

        # Track fields inherited from base models.
        inherited_attributes = set()
        # Do the appropriate setup for any model parents.
        for base in new_class.mro():
            if base not in parents or not hasattr(base, "_meta"):
                # Things without _meta aren't functional models, so they're
                # uninteresting parents.
                inherited_attributes.update(base.__dict__)
                continue

            parent_fields = base._meta.local_fields + base._meta.local_many_to_many
            if not base._meta.abstract:
                # Check for clashes between locally declared fields and those
                # on the base classes.
                for field in parent_fields:
                    if field.name in field_names:
                        raise FieldError(
                            "Local field %r in class %r clashes with field of "
                            "the same name from base class %r."
                            % (
                                field.name,
                                name,
                                base.__name__,
                            )
                        )
                    else:
                        inherited_attributes.add(field.name)

                # Concrete classes...
                base = base._meta.concrete_model

                # skipping ModelBase logic related to ptr fields for parent links

                # IMPORTANT: parents[base] entry is required to create a link between the class and parents that
                #            define fields. It is None JSONModel inheritance works more like regular OOP inheritance,
                #            where with Django Model classes, it can require joining across multiple tables (e.g in
                #            the case of a Model subclassing another non-abstract Model).
                new_class._meta.parents[base] = None
            else:
                base_parents = base._meta.parents.copy()

                # Add fields from abstract base class if it wasn't overridden.
                for field in parent_fields:
                    if (
                        field.name not in field_names
                        and field.name not in new_class.__dict__
                        and field.name not in inherited_attributes
                    ):
                        new_field = copy.deepcopy(field)
                        new_class.add_to_class(field.name, new_field)
                        # Replace parent links defined on this base by the new
                        # field. It will be appropriately resolved if required.
                        if field.one_to_one:
                            for parent, parent_link in base_parents.items():
                                if field == parent_link:
                                    base_parents[parent] = new_field

                # Pass any non-abstract parent classes onto child.
                new_class._meta.parents.update(base_parents)

            # Inherit private fields (like GenericForeignKey) from the parent
            # class
            for field in base._meta.private_fields:
                if field.name in field_names:
                    if not base._meta.abstract:
                        raise FieldError(
                            "Local field %r in class %r clashes with field of "
                            "the same name from base class %r."
                            % (
                                field.name,
                                name,
                                base.__name__,
                            )
                        )
                else:
                    field = copy.deepcopy(field)
                    if not base._meta.abstract:
                        field.mti_inherited = True
                    new_class.add_to_class(field.name, field)

        return new_class

    def add_to_class(cls, name, value):
        ModelBase.add_to_class(cls, name, value)

    def _prepare(cls):
        ModelBase._prepare(cls)


class JSONModel(metaclass=JSONModelBase):
    """
    Base class for models used with BaseJSONModelField.

    Adapted from Django's Model class.
    """

    _meta: JSONModelOptions
    _selector_value: Any = None

    def __init__(self, _has_input: bool = None, _skip_clean: bool = False, **data):
        """
        Initializes an instance either with an initial set of data that must include values for required fields, or as
        a completely empty instance (as is done for ModelForm instances used to create new objects)

        Arguments:
            _has_input - for internal use - allows the initializer to distinguish between being invoked with no data
                         (effectively an uninitialized instance), and being invoked with an empty data mapping
            _skip_clean - for internal use - when True, skips validation if input is provided. Used when initializing
                          objects from DB JSON data to prevent malformed data from breaking object loading.
        """

        if _has_input is None:
            _has_input = bool(data)

        if not _has_input:
            self._init_field_defaults()
            return

        self._init_fields(data)

        if not _skip_clean:
            # clean is only run if values are provided, otherwise it will throw validation errors for any required
            # fields and break forms for new objects
            self.full_clean()

    def _init_fields(self, initial: dict) -> None:
        """
        Iterates through all defined fields and sets a value from the `initial` dict, defaulting to None.

        This ensures that the DeferredAttribute accessor added to the model instance for accessing field values is
        replaced; if it is not replaced, it will raise an error if anything attempts to access the field's value.

        Parameters
        ----------
        initial
        """

        # Iterate through the list of fields rather than data keys so that a None value is set for any missing fields
        for field in self._meta.fields:
            value = initial[field.name] if field.name in initial else field.get_default()
            setattr(self, field.name, value)

    def _init_field_defaults(self) -> None:
        for field in self._meta.fields:
            if field.has_default():
                setattr(self, field.name, field.get_default())

    def _ensure_fields_initialized(self, data: dict = None) -> None:
        """
        Similar to `_init_fields`, ensures all fields have a value set to prevent errors when attempting to use their
        field accessor. Unlike _init_fields, only sets a None value if no value has been defined yet.
        """

        for field in self._meta.fields:
            try:
                getattr(self, field.name)
            except (FieldDoesNotExist, KeyError):
                # field value has not been set yet, still using DeferredAttribute
                setattr(self, field.name, data.get(field.name) if data else None)

    ##############################
    # adaptations of Model methods
    ##############################

    @classmethod
    def check(cls, **kwargs):
        # adapted from Model.check - most checks from Model.check aren't needed here since they deal with features not
        # supported for JSONModel
        errors = [
            *cls._check_fields(**kwargs),
        ]

        return errors

    @classmethod
    def _check_fields(cls, **kwargs):
        """Perform all field checks."""

        # adapted from Model._check_fields - only local fields need to be checked, no other types of fields are
        # supported

        errors = []
        for field in cls._meta.local_fields:
            errors.extend(field.check(**kwargs))
        return errors

    ###############################################################################################################
    # passthrough methods - these can be used as-is from Django's Model class directly rather than copy/pasting the
    # code
    ###############################################################################################################

    def full_clean(self, exclude=None, *args, **kwargs):
        self._ensure_fields_initialized()
        return Model.full_clean(self, exclude, validate_unique=False)

    def clean_fields(self, exclude=None):
        return Model.clean_fields(self, exclude)

    def _get_FIELD_display(self, field):
        # used by fields that allow "choices"
        return Model._get_FIELD_display(self, field)

    def clean(self):
        """
        Hook for doing any extra model-wide validation after clean() has been
        called on every field by self.clean_fields. Any ValidationError raised
        by this method will not be associated with a particular field; it will
        have a special-case association with the field defined by NON_FIELD_ERRORS.
        """
        pass

    ###################################################################################################################
    # placeholder methods - these are methods used by Django Models that pertain to functionality not needed or not
    #                       supported by JSONModel. They are provided as noops so to allow JSONModel classes to be used
    #                       interchangeably with Django Model classes for ModelForms and ModelAdmin.
    ###################################################################################################################

    def validate_unique(self, exclude=None):
        pass

    def refresh_from_db(self, using=None, fields=None):
        pass

    def clone(self):
        return self.__class__(**{field.name: getattr(self, field.name) for field in self._meta.fields})



from typing import Callable, Optional, Type, Union

from django.core.exceptions import ValidationError
from django.forms import BaseForm, BaseModelForm, BoundField, Field

from .widgets import NestedFormInput


class NestedFormBoundField(BoundField):
    """
    Subclasses the default BoundField to add support for initializing nested forms as a field's presentation.

    BoundFields are used to manage binding a field instance of a form to a specific data value since Field
    implementations on their own are stateless.
    """

    field: "NestedFormField"

    def get_nested_form_class(self) -> Optional[Type[BaseForm]]:
        return self.field.get_nested_form_class(self.form)

    def get_nested_form_kwargs(self, nested_form_class: Type[BaseForm]) -> dict:
        return self.field.get_nested_form_kwargs(self.form, self.name, nested_form_class)

    def get_nested_form(self) -> Optional[BaseForm]:
        nested_form_class = self.get_nested_form_class()

        if nested_form_class is None:
            return None

        nested_form_kwargs = self.get_nested_form_kwargs(nested_form_class)

        if "instance" in nested_form_kwargs and not issubclass(nested_form_class, BaseModelForm):
            nested_form_kwargs.pop("instance")

        return nested_form_class(**nested_form_kwargs)

    def as_widget(self, widget=None, attrs=None, only_initial=False):
        attrs = attrs or {}
        attrs["form"] = self.get_nested_form()
        return super().as_widget(widget, attrs, only_initial)


NestedFormClassFn = Callable[[Union[BaseForm, dict]], Optional[Type[BaseForm]]]
NestedFormClassArg = Union[Type[BaseForm], NestedFormClassFn]


class NestedFormField(Field):
    """
    A form field implementation that can be used to represent dictionary-like data using a nested Django form.
    """

    bound_field_class: Type[NestedFormBoundField] = NestedFormBoundField
    nested_form_class: Optional[Type[BaseForm]]
    widget: Type[NestedFormInput] = NestedFormInput

    _invalid_nested_field = "invalid_nested_field"

    def __init__(self, nested_form_class: NestedFormClassArg, *args, **kwargs):
        self._nested_form_class = nested_form_class
        if isinstance(nested_form_class, type) and issubclass(nested_form_class, BaseForm):
            self.nested_form_class = nested_form_class
        else:
            setattr(self, "get_nested_form_class", nested_form_class)

        # "required" validation will be delegated to the nested form
        kwargs.pop("required", None)
        super().__init__(required=False, *args, **kwargs)

        # check for BaseForm rather than Form - BaseForm is what is used by the modelform_factory function
        if isinstance(nested_form_class, type) and not issubclass(nested_form_class, BaseForm):
            raise TypeError("When specifying a type for nested_form_class, it must be a subclass of Form")

        if not isinstance(self.widget, NestedFormInput):
            raise TypeError("widgets for NestedFormField must be an instance of NestedFormInput or a subclass thereof")

        if not issubclass(self.bound_field_class, NestedFormBoundField):
            raise TypeError("bound_field_class must be a subclass of NestedFormBoundField")

        # TODO: set widget.is_required on widget based on whether the nested form has any required fields

    def get_nested_form_class(self, host: Union[BaseForm, dict], raw_data: dict = None) -> Optional[Type[BaseForm]]:
        return self.nested_form_class

    def clean(self, value):
        validation_form = self.get_form_for_validation(raw_data=value)
        if validation_form is None:
            setattr(self, "_validation_form_cleaned_data", value)
        else:
            validation_form.full_clean()
            setattr(self, "_validation_form_cleaned_data", validation_form.cleaned_data)
        return super().clean(value)

    def to_python(self, value):
        validation_form = self.get_form_for_validation(raw_data=value)
        if validation_form is None:
            return super().to_python(value)

        validation_form.is_valid()
        return validation_form.cleaned_data

    def get_form_for_validation(self, raw_data: dict = None, cleaned_data: dict = None):
        """
        Gets an instance of the nested form to be used for data conversion / cleaning / validation.
        """

        # This is a little naughty since it seems like field instances aren't supposed to be stateful or keep references
        # to anything specific to a form value, but it also seemed silly to have to instantiate the nested form and have
        # it run its cleaning / validation logic multiple times in order to maintain the statelessness.

        if raw_data is None and cleaned_data is None:
            raise ValueError("one of raw_data or cleaned_data must be specified")

        if (
            hasattr(self, "_validation_form")
            and (raw_data is not None and getattr(self, "_validation_form_raw_data", None) == raw_data)
            or (cleaned_data is not None and getattr(self, "_validation_form_cleaned_data", None) == cleaned_data)
        ):
            return getattr(self, "_validation_form")

        if raw_data is None:
            raise ValueError("raw_data must be provided when there is no previously initialized form")

        if cleaned_data is not None:
            # this shouldn't happen, something weird is going on
            raise ValueError("cleaned_data must not be provided when there is no previously initialized form")

        host_data = raw_data.pop(self.widget.included_from_host_key, None)
        nested_form_class = self.get_nested_form_class(host_data, raw_data=raw_data)
        form = nested_form_class(data=raw_data) if nested_form_class is not None else None

        setattr(self, "_validation_form_raw_data", raw_data)
        setattr(self, "_validation_form", form)
        return form

    def validate(self, value):
        super().validate(value)

        validation_form = self.get_form_for_validation(cleaned_data=value)

        if validation_form is not None and not validation_form.is_valid():
            # The empty string used for the validation message allows any validation errors on the nested form to bubble
            # up, and since there's no content for the message, the UI isn't affected and won't show seemingly
            # duplicative errors.
            raise ValidationError("", code=self._invalid_nested_field)

    def get_bound_field(self, form, field_name) -> BoundField:
        return self.bound_field_class(form, self, field_name)

    def get_nested_form_kwargs(self, host_form: BaseForm, field_name: str, nested_form_class: Type[BaseForm]) -> dict:
        return dict(
            prefix=host_form.add_prefix(field_name),
            data=self.get_nested_form_data(host_form, field_name, nested_form_class),
            initial=self.get_nested_form_initial(host_form, field_name, nested_form_class),
            renderer=host_form.renderer,
        )

    def get_nested_form_data(self, host_form: BaseForm, field_name: str, nested_form_class: Type[BaseForm]) -> dict:
        # form.data is defaulted to an empty dict in Form.__init__, so it will always have a non-None value regardless
        # of whether it is bound - however, Form.__init__ sets is_bound based on whether the data and files arguments
        # passed are None, so is_bound must be checked to prevent the nested form from thinking it is bound.
        return host_form.data.copy() if host_form.is_bound else None

    def get_nested_form_initial(self, host_form: BaseForm, field_name: str, nested_form_class: Type[BaseForm]) -> dict:
        return host_form.initial.pop(field_name, None)



from __future__ import annotations

from json import JSONEncoder
from typing import Any

from django_json_model_field.db.models import JSONModel
from . import constants
from .types import JSONClassWrapper, JSONModelDict


class JSONModelEncoder(JSONEncoder):
    """
    Custom JSONEncoder that can serialize JSONModel or JSONClassWrapper instances.

    This encoder is used by JSONModelDictField to support using JSONModels in dicts, which is not currently supported
    by JSONModelField.
    """

    def default(self, o: object):
        return collect_data(o)

    def encode(self, o):
        return super().encode(collect_data(o))


_COLLECTABLE_TYPES = (JSONModel, JSONClassWrapper, JSONModelDict, dict)


def collect_data(value: Any) -> Any:
    """
    Recursively normalizes nested JSONModel and JSONClassWrapper instances to dicts to avoid double-serializing nested
    objects.
    """

    # important to check this first so __json_class__ can be added
    if isinstance(value, JSONClassWrapper):
        return _collect_data_from_json_class_wrapper(value)

    if isinstance(value, JSONModelDict):
        db_json = value.as_db_json()
        if isinstance(db_json, JSONClassWrapper):
            return _collect_data_from_json_class_wrapper(db_json)
        return _collect_data_from_dict(db_json)

    if isinstance(value, JSONModel):
        return _collect_data_from_json_model(value)

    if isinstance(value, dict):
        return _collect_data_from_dict(value)

    return value


def _collect_data_from_json_model(instance: JSONModel) -> dict[str, Any]:

    field_value = {
        field.name: collect_data(field.value_from_object(instance))
        for field in instance._meta.get_fields()
    }

    model = type(instance)
    field_value[constants.JSON_CLASS_KEY] = f"{model.__module__}.{model.__name__}"
    return field_value


def _collect_data_from_json_class_wrapper(instance: JSONClassWrapper) -> dict[str, Any]:
    model = instance.wraps
    collected = dict(collect_data(instance.data))
    collected[constants.JSON_CLASS_KEY] = f"{model.__module__}.{model.__name__}"
    return collected


def _collect_data_from_dict(instance: dict[str, Any], *, use_json_class: bool = True) -> dict[str, Any] | None:
    if instance is None:
        return None

    collected = {key: collect_data(sub_value) for key, sub_value in instance.items()}

    if use_json_class is False:
        return collected

    model = type(instance)
    if model != dict:
        collected[constants.JSON_CLASS_KEY] = f"{model.__module__}.{model.__name__}"

    return collected



from __future__ import annotations

from typing import Any, Type

from django.core import checks
from django.db.models import JSONField

from django_json_model_field.serialization import JSONModelDecoder, JSONModelEncoder
from django_json_model_field.serialization.json_model_encoder import _collect_data_from_dict
from django_json_model_field.serialization.types import JSONModelDict


class JSONModelDictField(JSONField):
    """
    A JSONField implementation that supports storing JSONModel instances in a dict.

    The django_json_model library doesn't support dicts yet. This field adds basic support for allowing JSONModels to
    be used for serializing and deserializing, but does not provide any support for Django forms or admin.

    json_dict_model must be a dict, a dict subclass, or a class that implements as_db_json(self), which returns a dict
    or JSONClassWrapper instance. All dict values must be JSON serializable, or instances of JSONModel or
    JSONClassWrapper.

    dict data is passed to json_dict_model as kwargs.
    """

    # JSONModelDictField works by using customized JSON encoder and decoders. The encoder stores the class path for
    # serialized JSONModel and JSONClassWrapper instances, and the decoder uses this information to deserialize the
    # data. Since JSONEncoder does not provide a hook for customizing serialization of dictionaries, data from dict
    # subclasses are wrapped in a JSONClassWrapper instance, which forces the encoder's default(obj) method to be
    # called, allowing it to store the class path.
    #
    # Example:
    #    Saving:
    #       my_model.custom_dict = CustomDictInstance({"key": JSONModelInstance(name="value")})
    #       my_model.save()
    #           JSONModelDictField.get_prep_value(...) ->
    #               {"key": {"name": "value", "__json_class__": "module.JSONModelInstance"}}  # gets stored in DB
    #
    #    Loading:
    #       my_model.custom_dict <- {"key": {"name": "value", "__json_class__": "module.JSONModelInstance"}}  # from DB
    #       JSONModelDictField.from_db_value(...) -> CustomDictInstance(**{"key": CustomDictInstance(name="value")})
    #       my_model.custom_dict = CustomDictInstance({"key": "value"})

    def __init__(
        self,
        json_dict_model: Type[JSONModelDict | dict],
        verbose_name=None,
        name=None,
        null=True,
        blank=True,
        encoder=None,
        decoder=None,
        **kwargs
    ):
        encoder = encoder or JSONModelEncoder
        decoder = decoder or JSONModelDecoder
        super().__init__(
            verbose_name=verbose_name,
            name=name,
            encoder=encoder,
            decoder=decoder,
            null=null,
            blank=blank,
            **kwargs,
        )
        self.json_dict_model = json_dict_model

    def get_prep_value(self, value: Any):
        if isinstance(value, JSONModelDict):
            value = value.as_db_json()

        # the top level dict does not need the __json_class__ key since the type is known from the field
        value = _collect_data_from_dict(value, use_json_class=False)
        prep_value = super().get_prep_value(value)
        return prep_value

    def from_db_value(self, value, expression, connection):
        # deserialization is handled by the JSONField superclass through the decoder
        if isinstance(value, str):
            value = super().from_db_value(value, expression, connection)

        if value is None:
            return None

        if isinstance(value, self.json_dict_model):
            return value

        assert isinstance(value, dict)
        return self.json_dict_model(**value)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["json_dict_model"] = self.json_dict_model
        return name, path, args, kwargs

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_serialization(),
        ]

    def _check_serialization(self):
        errors: list[checks.CheckMessage] = []

        if not issubclass(self.encoder, JSONModelEncoder):
            errors.append(
                checks.Critical("JSONModelDictField encoder must be a subclass of JSONModelEncoder", obj=self)
            )

        if not issubclass(self.decoder, JSONModelDecoder):
            errors.append(
                checks.Critical("JSONModelDictField decoder must be a subclass of JSONModelDecoder", obj=self)
            )

        return errors



from typing import Optional, Tuple, Type

from django.db.models.constants import LOOKUP_SEP
from django.forms import Form, Widget

from django_json_model_field.util import get_type_init_with_metaclass


class NestedFormInput(Widget):
    """
    Widget implementation used to render a nested Django form as the UI representation of a field
    """

    @classmethod
    def with_fields_from_host(cls, *fields) -> Type["NestedFormInput"]:
        """
        A class factory method that allows specifying fields from the host model for that will be included with the data
        for the nested form when calling `value_from_datadict`. When specified, the values for these fields will be
        included in a `_host` key. This is used with ConditionalJSONModelField to help determine when the value of the
        selector field has changed.
        """

        return get_type_init_with_metaclass(cls)(cls.__name__, (cls,), {"include_from_host": tuple(fields)})

    template_name = "django_json_model_field/forms/widgets/nested_form_input.html"

    include_from_host: Tuple[str, ...] = ()
    included_from_host_key = "_host"

    def get_context(self, name, value, attrs):
        attrs = attrs or {}
        form: Optional[Form] = attrs.pop("form")

        context = super().get_context(name, value, attrs)
        context.update(nested_form=form)

        return context

    def value_from_datadict(self, data, files, name):
        """
        Used to extract data from a form POST to a dict by looking for data entries whose key is prefixed with the name
        of the widget's field.
        """

        # names for nested form fields will be prefixed with the name of the original field
        prefix = f"{name}-"
        nested_keys = [(key, key[len(prefix) :]) for key in data if key.startswith(prefix)]
        nested_data = {field: data[data_key] for data_key, field in nested_keys}

        if self.include_from_host:
            prefix_parts = name.split("-")
            host_prefix = "-".join(prefix_parts[:-1])
            host_prefix = f"{host_prefix}-" if host_prefix else ""
            host_keys = [
                (f"{host_prefix}{field}", field) for field in self.include_from_host if LOOKUP_SEP not in field
            ]
            host_data = {field: data[data_key] for data_key, field in host_keys}
            nested_data.update({self.included_from_host_key: host_data})

        return nested_data

    def value_omitted_from_data(self, data, files, name):
        """
        This method normally checks the posted data to see if there's a missing entry for the widget's field, but nested
        data is spread across multiple fields, so use value_from_datadict to see if any of the nested form's fields have
        entries in the data.
        """

        return self.value_from_datadict(data, files, name) is None



from typing import Any, Dict, List, Optional, Type, Union, cast

from django.core import checks
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import Field, Model
from django.db.models.constants import LOOKUP_SEP
from django.utils.functional import cached_property

from django_json_model_field.util import get_type_init_with_metaclass

from .base_json_model_field import BaseJSONModelField
from .conditional_json_model_attribute import ConditionalJSONModelAttribute
from .json_model import JSONModel

from django.utils.text import slugify


class ConditionalJSONModelField(BaseJSONModelField):
    @classmethod
    def _make_selector_json_model_class(
        cls, selector_field: str, selector_value: Any, json_model_class: Type[JSONModel]
    ) -> Type[JSONModel]:
        """
        Subclass each mapped JSONModel class to set the _selector_value attribute - this makes it easier to look up
        later in contexts where we have a reference to the class or an instance of the model, but not the host model or
        form to determine what the selector value is.
        """

        selector_field_title = selector_field.replace("_", " ").title().replace(" ", "")
        select_value_title = (
            slugify(str(selector_value).replace(".", "-")).replace("-", " ").replace("_", " ").title().replace(" ", "")
        )

        # use a more specific name to make debugging easier
        name = f"{json_model_class.__name__}For{selector_field_title}{select_value_title}"
        return get_type_init_with_metaclass(json_model_class)(
            name,
            (json_model_class,),
            {
                "__name__": name,
                "__module__": json_model_class.__module__,
                "__classcell__": getattr(json_model_class, "__classcell__", None),
                "_selector_value": str(selector_value),
            },
        )

    _selector_data_key = "_selector"

    def __init__(
        self,
        selector_field: str,
        model_map: Dict[Any, Type[JSONModel]],
        selector_data_key: str = None,
        verbose_name: str = None,
        name: str = None,
        null: bool = True,
        blank: bool = True,
        encoder=None,
        decoder=None,
        **kwargs,
    ):
        """

        Parameters
        ----------
        selector_field The name of the field used to determine which entry from model_map will be used to work with
                       and manipulate the JSON data stored using the field. Note that values accessed from
                       `selector_field` will be stringified before looking up the entry in model_map.
        model_map      A map of possible values from selector_field to the JSONModel that will be used to represent
                       the data. Note that keys will be stringified before being used for lookups. Selector values
                       with missing entries are assumed to have no model and have no data requirements.
                       TODO: add an option to raise a validation or check error for missing entries?
        selector_data_key Use an alternate key to store the selector value used to determine the `JSONModel` class
                          used to manage the JSON data. The default is "_selector".

        """
        # validation of model field arguments is handled by Django's checks system - see the `check` method below.

        self.selector_field_name = selector_field
        self.model_map = {
            str(key): self._make_selector_json_model_class(selector_field, key, value)
            for key, value in model_map.items()
        }

        # make a separate copy of model_map for deconstruction - the original JSONModel classes must be used here since
        # deconstruct is used to generate migrations, and the type returned by make_selector_json_model_class will not
        # get created until the field is initialized at runtime
        self._model_map = {str(key): value for key, value in model_map.items()}

        if selector_data_key:
            self._selector_data_key = selector_data_key

        super().__init__(
            verbose_name=verbose_name,
            name=name,
            encoder=encoder,
            decoder=decoder,
            null=null,
            blank=blank,
            **kwargs,
        )

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        args = [self.selector_field_name, self._model_map, *args]
        kwargs.update(selector_data_key=self._selector_data_key)
        return name, path, args, kwargs

    def get_formfield_kwargs(self, **kwargs):
        kwargs = super().get_formfield_kwargs(**kwargs)
        kwargs.update(
            nested_model_class=self.get_json_model_class,
            selector_field=self.selector_field_name,
            get_selector_value=self._get_selector_value
        )
        return kwargs

    def get_json_model_host_data(self, data: Optional[dict]) -> dict:
        return data.pop(self._selector_data_key, {}) if data else {}

    def get_selector_value(self, host: Union[Model, dict]) -> Optional[Any]:
        """
        Attempts to retrieve the value of the selector_field from the specified host model or data dictionary.
        """

        if host is None or host == {}:
            return None

        value = self._get_selector_value(host) if isinstance(host, Model) else host.get(self.selector_field_name)
        return str(value) if value is not None else None

    def _get_selector_value(self, host: Model) -> Optional[Any]:
        value: Any = host
        parts = self.selector_field_path.copy()
        while len(parts):
            part = parts.pop(0)
            value = getattr(value, part)

        return value

    @staticmethod
    def _get_field(model_class: Type[Model], field_name: str):
        return next((field for field in model_class._meta.fields if field.name == field_name), None)

    def get_json_model_class(self, host: Union[Model, dict]) -> Optional[Type[JSONModel]]:
        """
        Returns the `JSONModel` to be used to represent the field data, if one can be determined either from json_model
        or selector_field and model_map.
        """

        selector = self.get_selector_value(host)
        if selector is None:
            return None
        return self.model_map.get(selector)

    def _get_db_prep_value_from_model(self, instance: "JSONModel", connection: BaseDatabaseWrapper) -> dict:
        data = super()._get_db_prep_value_from_model(instance, connection)
        if instance._selector_value:
            data.update({self._selector_data_key: {self.selector_field_name: instance._selector_value}})
        return data

    @cached_property
    def selector_field_path(self) -> List[str]:
        """
        A list of segments representing the full path to the selector field through any relation lookups
        """

        return self.selector_field_name.split(LOOKUP_SEP)

    @cached_property
    def selector_field(self) -> Field:
        """
        Returns a reference to the Field instance represented by selector_field_path.

        Raises an Exception if the field cannot be found, though in practice, this would be caught by the field checks
        in _check_json_model_arguments.
        """

        selector_field = self._get_selector_field()
        if selector_field:
            return selector_field

        raise Exception(f"Could not locate selector field {self.selector_field_name} on model {self.model.__name__}")

    def contribute_to_class(self, cls, name, private_only=False):
        super().contribute_to_class(cls, name, private_only)

        setattr(cls, self.attname, ConditionalJSONModelAttribute(self))

    def _get_selector_field(self) -> Optional[Field]:
        """
        Attempts to find the Field instance identified by selector_field_path. If the field cannot be found, None is
        returned.
        """

        part_field = self
        parts = self.selector_field_path.copy()
        while len(parts):
            if not part_field:
                return None
            host_model = part_field.related_model if part_field.is_relation else part_field.model
            part = parts.pop(0)
            part_field = self._get_field(host_model, part)

        return part_field

    def _check_json_model_arguments(self, **kwargs):
        """
        Checks arguments related to specifying the JSONModel for the field to ensure the right combination of arguments
        was provided, and that the arguments are of valid types.
        """

        if self.selector_field_name is None and self.model_map is None:
            return [checks.Critical("Both selector_field and model_map must have a value", obj=self)]

        if self.selector_field_name is None:
            return [
                checks.Critical(
                    (
                        "selector_field must be set to the name of another field on the model, or a lookup to a field "
                        "on a relation"
                    ),
                    obj=self,
                )
            ]

        if self.model_map is None:
            return [
                checks.Critical(
                    "model_map is required",
                    hint="model_map must be set to a dict mapping of possible selector values to JSONField class types",
                    obj=self,
                )
            ]

        if self.model_map == {}:
            return [
                checks.Critical(
                    "model_map must have a non-empty value",
                    hint="model_map must be set to a dict mapping of possible selector values to JSONField class types",
                    obj=self,
                )
            ]

        selector_field = self._get_selector_field()
        if selector_field is None:
            return [
                checks.Critical(
                    f"Invalid selector_field: No field named {self.selector_field_name}",
                    obj=self,
                )
            ]

        if selector_field.is_relation:
            return [
                checks.Critical(
                    "selector_field cannot be a relation field",
                    obj=self,
                )
            ]

        # cast to make mypy happy - previous checks have already established that model_map is not None
        model_map = cast(Dict[str, Type[JSONModel]], self.model_map)
        errors = []
        for selector_value, json_model in model_map.items():
            errors.extend(self._check_model_argument(json_model, f"model_map[{selector_value}]"))

        return errors or self._check_json_models(**kwargs)

    def _check_json_models(self, **kwargs):
        errors = []
        for json_model in self.model_map.values():
            errors.extend(json_model.check(**kwargs))

        return errors



from typing import Optional, TYPE_CHECKING

from .json_model import JSONModel

if TYPE_CHECKING:
    from .conditional_json_model_field import ConditionalJSONModelField


class ConditionalJSONModelAttribute:
    """
    A custom accessor to be used with ConditionalJSONModelField.

    ConditionalJSONModelField creates subclasses of each JSONModel class defined in the model_map, each of which have
    their respective selector value set on the _selector_value attribute. When JSONModel values are saved to their
    JSON column in a database, ConditionalJSONModelField adds a structure to the JSON that includes the value of
    _selector_value so that the JSON structure can be deserialized into the correct JSONModel class when it is read from
    the database. The _selector_value attribute allows this to be done without having to make additional checks to the
    actual selector field, which can potentially incur additional performance or database access overhead if the
    selector field includes a relation lookup.

    However, if there is user code that creates its own instances of a JSONModel, it will likely not be using the
    _selector_value-annotated classes, but the original definition of the class that will not have a value set for
    _selector_value. In these cases, when that object is set on a ConditionalJSONModelField, this accessor will
    determine the expected JSONModel class based on the value of the selector field, and set the attribute value on the
    user-created instance. Note that this is specifically done on the instance and not the original JSONModel class
    definition because it is possible that JSONModels can be shared between multiple fields.
    """

    def __init__(self, field: "ConditionalJSONModelField"):
        self.field = field

    def _get_value(self, host_instance):
        return host_instance.__dict__.get(self.field.attname)

    def _set_value(self, host_instance, value):
        # store the value on the host instance's __dict__, otherwise the field will be considered to be a "deferred"
        # field by the logic in Model.save(), and the field will be omitted from updates to the DB
        host_instance.__dict__[self.field.attname] = value

    def __get__(self, host_instance, owner):
        return self._get_value(host_instance)

    def __set__(self, host_instance, value: Optional[JSONModel]):
        if value is None or value == {}:
            self._set_value(host_instance, value)
            return

        # _make_selector_json_model_class adds the _selector_value attribute when mapping JSONModel classes - if the
        # JSONModel value already has a value set for _selector_value, just store the value and don't bother with the
        # checks below. This can potentially avoid performance penalties if the selector field uses one or more relation
        # lookups.
        if bool(getattr(value, "_selector_value", None)):
            self._set_value(host_instance, value)
            return

        json_model_class = type(value)

        # this will be a subclass of the originally defined JSONModel class due to _make_selector_json_model_class
        expected_json_model_class: Optional[type] = self.field.get_json_model_class(host_instance)

        if expected_json_model_class is None:
            # No JSONModel value is needed based on the current selector value - it should not be getting set to a
            # non-None value
            raise TypeError(f"{self.field.attname} is not expecting a value")

        # for a valid value, it will be:
        #   - the same class if the value is being set from a JSONModelField, or any source that uses
        #     get_json_model_class to get the correct JSONModel class for the host host_instance's state
        #   - the immediate superclass of expected_json_model_class if the value being set is an instance created by
        #     user code that imported the JSONModel class directly from its declaration module
        if not issubclass(expected_json_model_class, json_model_class):
            expected_mro = expected_json_model_class.mro()
            # [0] is the class itself, [1] would be its immediate superclass
            original_json_class = expected_mro[1]
            raise TypeError(f"{self.field.attname} must be an instance of {original_json_class.__name__}")

        # _make_selector_json_model_class adds the _selector_value attribute when mapping JSONModel classes, but
        # it will not be set on user-created instances - set it on the instance here (NOT on json_model_class, it
        # is possible to reuse JSONModels between multiple fields!)
        setattr(value, "_selector_value", getattr(expected_json_model_class, "_selector_value"))
        self._set_value(host_instance, value)








from importlib import import_module
from json import JSONDecoder

from django.db import connection

from django_json_model_field.db.models import JSONModel
from . import constants


class JSONModelDecoder(JSONDecoder):

    def __init__(self, **kwargs):
        super().__init__(object_hook=self._object_hook, **kwargs)

    @classmethod
    def _object_hook(cls, d: dict):
        if constants.JSON_CLASS_KEY not in d:
            return d

        # load the model class from __json_class__ and instantiate it
        model_name = d.pop(constants.JSON_CLASS_KEY)
        module_name, model_name = model_name.rsplit(".", 1)
        model_module = import_module(module_name)
        model = getattr(model_module, model_name)

        if issubclass(model, JSONModel):
            from django_json_model_field.db.models import JSONModelField

            field = JSONModelField(model)
            return field.from_db_data_dict(d, connection)

        return model(**d)
